import json
import logging
import requests

from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation as Imp
from rest.connector.utils import get_token

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Imp):
    '''Rest Implementation for Webex

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for Webex

    YAML Example
    ------------

        devices:
            webex:
                os: webex
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "10.1.1.1"
                        credentials:
                            rest:
                                token: <webexaccesstoken>

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('testbed.yaml')
        >>> device = testbed.devices['webex']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''
    @BaseConnection.locked
    def connect(self, timeout=30):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.


        YAML Example
        ------------

            devices:
                webex:
                    os: webex
                    connections:
                        rest:
                            class: rest.connector.Rest
                            ip : "10.1.1.1"
                            credentials:
                                rest:
                                    token: <webexaccesstoken>

        Code Example
        ------------

            >>> from pyats.topology import loader
            >>> testbed = loader.load('testbed.yaml')
            >>> device = testbed.devices['webex']
            >>> device.connect(alias='rest', via='rest')
        '''

        if self.connected:
            return

        if 'host' in self.connection_info:
            ip = self.connection_info['host']
        else:
            ip = self.connection_info['ip'].exploded
        if 'port' in self.connection_info:
            port = self.connection_info['port']
            self.url = 'https://{ip}:{port}/'.format(ip=ip, port=port)
        else:
            self.url = 'https://{ip}/'.format(ip=ip)
        login_url = '{f}v1/people/me'.format(f=self.url)

        self.token = get_token(self)

        self.headers = {
            'Authorization': 'Bearer {}'.format(self.token),
        }

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.session = requests.Session()

        # Connect to the device via requests
        response = self.session.get(login_url, timeout=timeout, \
            headers=self.headers)
        log.info(response)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'"\
                                        .format(ip=ip, c=response.status_code,
                                                ok=requests.codes.ok))
        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def isconnected(func):
        '''Decorator to make sure session to device is active

           There is limitation on the amount of time the session cab be active
           on the APIC. However, there are no way to verify if
           session is still active unless sending a command. So, its just
           faster to reconnect every time.
         '''
        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()
            finally:
                ret = func(self, *args, **kwargs)
            return ret

        return decorated

    @BaseConnection.locked
    def _request(self, method, dn, **kwargs):
        """ Wrapper to send REST command to device

        Args:
            method (str): session request method

            dn (str): rest endpoint

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok

        """
        if not self.connected:
            raise Exception("'{d}' is not connected for alias '{a}'".format(
                d=self.device.name, a=self.alias))

        # Deal with the dn
        full_url = '{f}{dn}'.format(f=self.url, dn=dn)

        if 'data' in kwargs:
            p = kwargs['data']
        elif 'json' in kwargs:
            p = kwargs['json']
        else:
            p = ''

        expected_return_code = kwargs.pop('expected_return_code', None)

        log.info("Sending {method} command to '{d}':"
                 "\nDN: {furl}\nPayload:{p}".format(method=method,
                                                    d=self.device.name,
                                                    furl=full_url,
                                                    p=p))

        # Send to the device
        response = self.session.request(method=method, url=full_url, **kwargs)

        # An expected return code was provided. Ensure the response has this code.
        if expected_return_code:
            if response.status_code != expected_return_code:
                raise RequestException(
                    "'{c}' result code has been returned for '{d}'.\n"
                    "Expected '{expected_c}' result code.\n"
                    "Response from server: {r}".format(
                        c=response.status_code,
                        d=self.device.name,
                        expected_c=expected_return_code,
                        r=response.text))
        else:
            # No expected return code provided. Make sure it was successful.
            try:
                response.raise_for_status()
            except Exception:
                raise RequestException("'{c}' result code has been returned "
                                       "for '{d}'.\nResponse from server: "
                                       "{r}".format(d=self.device.name,
                                                    c=response.status_code,
                                                    r=response.text))

        log.info("Response from '{dev}':\n"
                 "Result Code: {c}\n"
                 "Response: {r}".format(dev=self.device.name,
                                        c=response.status_code,
                                        r=response.text))

        # In case the response cannot be decoded into json
        # warn and return the raw text
        if response.text:
            try:
                output = response.json()
            except Exception:
                log.warning('Could not decode json. Returning text!')
                output = response.text
        else:
            output = response.text

        return output

    @BaseConnection.locked
    @isconnected
    def get(self, dn, headers=None, timeout=30, **kwargs):
        """ GET REST Command to retrieve information from the device

        Args:
            dn (str): Unique distinguished name that describes the object
                      and its place in the tree.

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time to allow rest call to return

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        if not headers:
            headers = self.headers
        if headers and 'Authorization' not in headers:
            headers.update({'Authorization': 'Bearer {}'.format(self.token)})

        return self._request('GET',
                             dn,
                             headers=headers,
                             timeout=timeout,
                             **kwargs)

    @BaseConnection.locked
    @isconnected
    def post(self, dn, payload, headers=None, timeout=30, **kwargs):
        """POST REST Command to configure new information on the device

        Args:
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            payload (dict): Dictionary containing the information to send via
                            the post

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        if not headers:
            headers = self.headers
        if headers and 'Authorization' not in headers:
            headers.update({'Authorization': 'Bearer {}'.format(self.token)})

        if isinstance(payload, str):
            payload = json.loads(payload)

        return self._request('POST',
                             dn,
                             data=payload,
                             headers=headers,
                             timeout=timeout,
                             **kwargs)

    @BaseConnection.locked
    @isconnected
    def delete(self, dn, headers=None, timeout=30, **kwargs):
        """DELETE REST Command to delete information from the device

        Args
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        if not headers:
            headers = self.headers
        if headers and 'Authorization' not in headers:
            headers.update({'Authorization': 'Bearer {}'.format(self.token)})

        return self._request('DELETE',
                             dn,
                             headers=headers,
                             timeout=timeout,
                             **kwargs)

    @BaseConnection.locked
    @isconnected
    def put(self, dn, payload, headers=None, timeout=30, **kwargs):
        """PUT REST Command to update existing information on the device

        Args
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            payload (dict): Dictionary containing the information to send via
                            the post

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        if not headers:
            headers = self.headers
        if headers and 'Authorization' not in headers:
            headers.update({'Authorization': 'Bearer {}'.format(self.token)})

        if type(payload) == str:
            payload = json.loads(payload)

        return self._request('PUT',
                             dn,
                             data=payload,
                             headers=headers,
                             timeout=timeout,
                             **kwargs)



""" Utilities shared by all plugin libraries. """
from pkg_resources import get_distribution, DistributionNotFound
import re
import requests
import subprocess

try:
    from pyats.utils.secret_strings import to_plaintext
except ImportError:
    def to_plaintext(string):
        return(str(string))


def get_username_password(connection):
    username = password = None
    if connection.connection_info.get('credentials'):
        try:
            username = str(
                connection.connection_info['credentials']['rest']['username'])
            password = to_plaintext(
                connection.connection_info['credentials']['rest']['password'])
        except Exception:
            pass

    if not username:
        username = connection.connection_info.get('username', \
            connection.device.tacacs.username \
            if connection.device.tacacs.username \
            else 'admin')

    if not password:
        password = connection.connection_info.get('password', \
            connection.device.passwords.tacacs \
            if connection.device.passwords.tacacs \
            else 'admin')

    return (username, password)


def get_token(connection):
    token = None
    if connection.connection_info.get('credentials'):
        try:
            token = str(
                connection.connection_info['credentials']['rest']['token'])
        except Exception:
            pass

    return token


def get_apic_sdk_version(ip):
    """
    :param ip: IP or hostname of the APIC controller
    :return: version of the APIC software as noted in the /cisco page
    """
    version = None
    url = f'https://{ip}/cobra'

    response = requests.get(url, verify=False)

    if response.status_code == requests.codes.ok:
        match = re.search(r'<title>.*\s([\d.]+)\s.*</title>', response.text)
        if match:
            version = match.groups()[0]

    return version


def get_installed_lib_versions(packages=('acicobra', 'acimodel')):
    """
    :param packages: (tuple) a list of packages to be checked
    :return: dictionary with packages as keys and versions as values
    """
    if not isinstance(packages, tuple) and not isinstance(packages, list):
        packages = [packages]
    versions = {}
    for package in packages:
        try:
            versions[package] = get_distribution(package).version
        except DistributionNotFound:
            versions[package] = None

    return versions


def get_file_links(ip, packages=('acicobra', 'acimodel')):
    """
    :param ip: IP or hostname of the APIC controller
    :param packages: (tuple) a list of packages to be downloaded
    :return: dictionary containing package names as keys and full path to files
    """
    if not isinstance(packages, list):
        packages = [packages]
    url = f'https://{ip}/cobra/_downloads'
    links_dict = {}

    r = requests.get(url, verify=False)
    files = re.findall(r'<a href=([\w\d.-]+.whl)>', r.text)

    for package in packages:
        for file in files:
            if package in file:
                links_dict[package] = f'{url}/{file}'

    return links_dict


def pip_install_from_link(ip, link=''):
    """
    :param ip: IP or hostname of the APIC controller
    :param link: link to file (https://{ip}/cobra/_downloads/{file_name})
    :return: True if install succeeded, False otherwise
    """
    params = ['pip', 'install', f'--trusted-host={ip}', link]

    r = subprocess.run(params)
    if r.returncode:
        return False
    return True


def verify_apic_version(ip):
    apic_version = get_apic_sdk_version(ip=ip)
    installed = get_installed_lib_versions(packages=['acicobra', 'acimodel'])

    if apic_version and any(map(lambda x: x != apic_version, installed)):
        links_dict = get_file_links(ip=ip, packages=[k for k, v in
                                                     installed.items()
                                                     if v != apic_version])
        for package in links_dict:
            if not pip_install_from_link(ip=ip, link=links_dict[package]):
                raise RuntimeError('There seems to be a problem with the '
                                   f'{package} automated installation.\n'
                                   'Please install packages manually from APIC')



import requests
import urllib3

from functools import wraps
from importlib import import_module
from logging import getLogger
from pprint import pformat
from re import search
from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.utils import get_username_password, verify_apic_version

# create a logger for this module
log = getLogger(__name__)


class AciCobra(BaseConnection):
    """ACI SDK (Cobra) Implementation for APIC

    Implementation of ACI SDK connection to devices
    based on pyATS BaseConnection for APIC

    YAML Example
    ------------

        devices:
            apic1:
                connections:
                    cobra:
                        class: rest.connector.Acisdk
                        ip : "2.3.4.5"
                credentials:
                    cobra:
                        username: admin
                        password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
        >>> device = testbed.devices['apic1']
        >>> device.connect(alias='cobra', via='cobra')
        >>> device.rest.connected
        True
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mo_dir = None

        # remove warnings for insecure HTTPS
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self._is_connected = False

        if 'host' in self.connection_info:
            ip = self.connection_info['host']
        else:
            ip = self.connection_info['ip'].exploded
        if 'port' in self.connection_info:
            port = self.connection_info['port']
            self.url = f'https://{ip}:{port}/'
        else:
            self.url = f'https://{ip}/'

        verify_apic_version(ip)

    @property
    def connected(self):
        """Is a device connected"""
        return self._is_connected

    @BaseConnection.locked
    def connect(self, timeout=90):
        """
        Connect to the device via ACI SDK
        Arguments
        ---------
            timeout (int): Timeout value
        Raises
        ------
        Exception
        ---------
            If the connection did not go well
        Note
        ----
        There is no return from this method. If something goes wrong, an
        exception will be raised.
        YAML Example
        ------------
            devices:
                apic1:
                    connections:
                        cobra:
                            class: rest.connector.Acisdk
                            ip : "2.3.4.5"
                    credentials:
                        cobra:
                            username: admin
                            password: cisco123
        Code Example
        ------------
            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
            >>> device = testbed.devices['apic1']
            >>> device.connect(alias='cobra', via='cobra')
        """

        if self.connected:
            return

        username, password = get_username_password(self)

        login_session = getattr(import_module('cobra.mit.session'), 'LoginSession')
        session = login_session(self.url, username, password, timeout=timeout)

        mo_directory = getattr(import_module('cobra.mit.access'), 'MoDirectory')
        self.mo_dir = mo_directory(session)

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.mo_dir.login()

        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    def disconnect(self):
        """disconnect the device for this particular alias"""

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.mo_dir.logout()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def is_connected(func):
        """Decorator to make sure session to device is active
           There is limitation on the amount of time the session cab be active
           on the APIC. However, there are no way to verify if
           session is still active unless sending a command. So, its just
           faster to reconnect every time.
        """
        @wraps(func)
        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                ret = func(self, *args, **kwargs)
            except Exception as e:
                if not hasattr(e, 'reason') or e.reason != 'Token was invalid (Error: Token timeout)':
                    raise

                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()

                ret = func(self, *args, **kwargs)
            return ret

        return decorated

    def log_action(func):
        """Decorator to log the actions made by MoDirectory functions.
        """
        @wraps(func)
        def decorated(self, *args, **kwargs):

            ret = func(self, *args, **kwargs)
            log.info(f'Output of function {func.__name__} is:\n{pformat(ret, indent=2)}')
            return ret

        return decorated

    @BaseConnection.locked
    @is_connected
    @log_action
    def query(self, queryObject):
        """
        Mimics same-name function from MoDirectory class (cobra.mit.access):
        Queries the MIT for a specified object. The queryObject provides a
        variety of search options.
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        return self.mo_dir.query(queryObject)

    @BaseConnection.locked
    @is_connected
    @log_action
    def commit(self, configObject, sync_wait_timeout=None):
        """
        Mimics same-name function from MoDirectory class (cobra.mit.access):
        Short-form commit operation for a configRequest
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        return self.mo_dir.commit(configObject, sync_wait_timeout=sync_wait_timeout)

    @BaseConnection.locked
    @is_connected
    @log_action
    def lookupByDn(self, dnStrOrDn, **queryParams):
        """
        Mimics same-name function from MoDirectory class (cobra.mit.access):
        A short-form managed object (MO) query using the distinguished name(Dn)
        of the MO.

        Args:
          dnStrOrDn:   dn of the object to lookup
          queryParams: a dictionary including the properties to the
            added to the query.
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        return self.mo_dir.lookupByDn(dnStrOrDn, **queryParams)

    @BaseConnection.locked
    @is_connected
    @log_action
    def lookupByClass(self, classNames, parentDn=None, **queryParams):
        """
        Mimics same-name function from MoDirectory class (cobra.mit.access):
        A short-form managed object (MO) query by class.

        Args:
          classNames: Name of the class to lookup
          parentDn:   dn of the root object were to start search from (optional)
          queryParams: a dictionary including the properties to the
            added to the query.
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        return self.mo_dir.lookupByClass(classNames, parentDn=parentDn, **queryParams)

    @BaseConnection.locked
    @is_connected
    @log_action
    def exists(self, dnStrOrDn):
        """
        Mimics same-name function from MoDirectory class (cobra.mit.access):
        Checks if managed object (MO) with given distinguished name (dn) is present or not

        Args:
          dnStrOrDn (str or cobra.mit.naming.Dn): A distinguished name as a
            :class:`cobra.mit.naming.Dn` or string

        Returns:
          bool: True, if MO is present, else False.
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        return self.mo_dir.exists(dnStrOrDn)

    @log_action
    def get_model(self, model):
        """
        Automatically import the required library and return the model class
        :param model: must contain module and class of desired model (eg. fv.Tenant)
        :return: requested Cobra model
        """
        module = None
        attribute = None

        if '.' in model:
            module, attribute = model.rsplit('.', 1)
        else:
            match = search(r'([a-z]*)([A-Z]\w*)', model)
            if match:
                module, attribute = match.groups()

        if module and attribute:
            return getattr(import_module(f'cobra.model.{module}'), attribute)
        else:
            raise NameError("'model' must contain <module>.<class of object>")

    def create(self, model, parent_mo_or_dn, **extra_parms):
        """
        Automatically import the required library and instantiate the model object
        :param model: must contain module and class of desired model (eg. fv.Tenant)
        :param parent_mo_or_dn: equivalent of parentMoOrDn for cobra models
        :return: requested Cobra model object, initialized with given parameters
        """
        return self.get_model(model)(parentMoOrDn=parent_mo_or_dn, **extra_parms)

    @BaseConnection.locked
    @is_connected
    @log_action
    def config_and_commit(self, mo, sync_wait_timeout=None,
                          expected_status_code=requests.codes.ok):
        """
        Add MO to ConfigRequest and push it to device
        :param mo:
        :param sync_wait_timeout:
        :param expected_status_code:
        :return:
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        config = getattr(import_module(f'cobra.mit.request'), 'ConfigRequest')()
        config.addMo(mo)
        resp = self.mo_dir.commit(configObject=config, sync_wait_timeout=sync_wait_timeout)
        if resp.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}', got:\n {msg}"\
                                   .format(d=self.device.name,
                                           c=resp.status_code,
                                           e=expected_status_code,
                                           msg=resp.text))
        return True



import json
import logging
import requests
import time
import urllib.parse

from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation as Imp
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Imp):
    """Rest Implementation for ND
    Implementation of Rest connection to devices based on pyATS BaseConnection
    for ND

    YAML Example
    ------------

        devices:
            nd1:
                os: nd
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                credentials:
                    rest:
                        username: admin
                        password: cisco123

    Code Example
    ------------
        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
        >>> device = testbed.devices['nd1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    """

    @BaseConnection.locked
    def connect(self, timeout=30, retries=3, retry_wait=10):
        """connect to the device via REST
        Arguments
        ---------
            timeout (int): Timeout value
            retries (int): Max retries on request exception (default: 3)
            retry_wait (int): Seconds to wait before retry (default: 10)

        Raises
        ------

        Exception
        ---------
            If the connection did not go well

        Note
        ----
        There is no return from this method. If something goes wrong, an
        exception will be raised.

        YAML Example
        ------------
            devices:
                nd1:
                    os: nd
                    connections:
                        rest:
                            class: rest.connector.Rest
                            ip : "2.3.4.5"
                    credentials:
                        rest:
                            username: admin
                            password: cisco123

        Code Example
        ------------
            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
            >>> device = testbed.devices['nd1']
            >>> device.connect(alias='rest', via='rest')
        """

        if self.connected:
            return

        if 'host' in self.connection_info:
            ip = self.connection_info['host']
        else:
            ip = self.connection_info['ip'].exploded
        if 'port' in self.connection_info:
            port = self.connection_info['port']
            self.url = 'https://{ip}:{port}/'.format(ip=ip, port=port)
        else:
            self.url = 'https://{ip}/'.format(ip=ip)
        login_url = '{f}login'.format(f=self.url)

        username, password = get_username_password(self)

        payload = {
            "userName": username,
            "userPasswd": password,
            "domain": "DefaultAuth"
        }

        headers = {
            'Content-Type': 'application/json'
        }

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.session = requests.Session()
        _data = json.dumps(payload)

        for _ in range(retries):
            try:
                # Connect to the device via requests
                response = self.session.post(login_url, data=_data, timeout=timeout,
                                             verify=False, headers=headers)
                log.info(response)

                # Make sure it returned requests.codes.ok
                if response.status_code != requests.codes.ok:
                    # Something bad happened
                    raise RequestException("Connection to '{ip}' has returned the "
                                           "following code '{c}', instead of the "
                                           "expected status code '{ok}'"
                                           .format(ip=ip, c=response.status_code,
                                                   ok=requests.codes.ok))
                break
            except Exception:
                log.warning('Request to {} failed. Waiting {} seconds before retrying\n'.format(
                    self.device.name, retry_wait), exc_info=True)
                time.sleep(retry_wait)
        else:
            raise ConnectionError('Connection to {} failed'.format(self.device.name))

        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def disconnect(self):
        """disconnect the device for this particular alias"""

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def isconnected(func):
        """Decorator to make sure session to device is active

           There is limitation on the amount of time the session can be active
           on the ND. However, there is no way to verify if
           session is still active unless sending a command. So, it's just
           faster to reconnect every time.
         """

        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()
            finally:
                ret = func(self, *args, **kwargs)
            return ret

        return decorated

    @BaseConnection.locked
    @isconnected
    def get(self, api_url, expected_status_code=requests.codes.ok,
            timeout=30, retries=3, retry_wait=10):
        """GET REST Command to retrieve information from the device
        Arguments
        ---------
            api_url (string): subdirectory part of the API URL
            expected_status_code (int): Expected result
            timeout (int): Maximum time
            retries (int): Number of retries in case of transmission error
            retry_wait (int): Seconds to wait between retries
        """

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        #Eliminate the starting "/" if present, as it may cause problems
        api_url = api_url.lstrip('/')
        # Deal with the url
        full_url = "{f}{api_url}".format(f=self.url, api_url=api_url)

        log.info("Sending GET command to '{d}':" \
                 "\nURL: {furl}".format(d=self.device.name, furl=full_url))

        for _ in range(retries):
            try:
                response = self.session.get(full_url, timeout=timeout, verify=False)
                break
            except Exception:
                log.warning('Request to {} failed. Waiting {} seconds before retrying\n'.format(
                    self.device.name, retry_wait), exc_info=True)
                time.sleep(retry_wait)
        else:
            raise RequestException('Sending "{furl}" to "{d}" has failed '
                                   'after {tries} tries'.format(furl=full_url,
                                                                d=self.device.name,
                                                                tries=retries))

        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=
            json.dumps(output, indent=2, sort_keys=True)))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("Sending '{furl} to '{d} has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{e}'"
                                   "'{e}'".format(furl=full_url,
                                                  d=self.device.name,
                                                  c=response.status_code,
                                                  e=expected_status_code))
        return output

    @BaseConnection.locked
    @isconnected
    def post(self, api_url, payload, expected_status_code=requests.codes.ok,
             content_type='json', timeout=30, retries=3, retry_wait=10):
        """POST REST Command to configure information from the device
        Arguments
        ---------
            api_url (string): subdirectory part of the API URL
            payload (dict|string): Information to send via the post
            expected_status_code (int): Expected result
            content_type(string): json / xml / form
            timeout (int): Maximum time
        """

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        # Eliminate the starting "/" if present, as it may cause problems
        api_url = api_url.lstrip('/')
        # Deal with the url
        full_url = '{f}{api_url}'.format(f=self.url, api_url=api_url)

        log.info("Sending POST command to '{d}':" \
                 "\nURL: {furl}\nPayload:{p}".format(d=self.device.name,
                                                     furl=full_url,
                                                     p=payload))
        headers = {'form': 'application/x-www-form-urlencoded',
                   'json': 'application/json',
                   'xml': 'application/xml'}

        if content_type == 'form' and isinstance(payload, dict):
            payload = urllib.parse.urlencode(payload, safe=':!')
        elif content_type == 'xml':
            if isinstance(payload, dict):
                raise ValueError("Error on {d} during POST command: "
                                 "Payload needs to be string in xml format if used "
                                 "in conjunction with content_type='xml' argument"
                                 .format(d=self.device.name))

        for _ in range(retries):
            try:
                # Send to the device
                if isinstance(payload, dict):
                    response = self.session.post(full_url, json=payload, timeout=timeout,
                                                 verify=False)
                else:
                    response = self.session.post(full_url, data=payload, timeout=timeout,
                                                 verify=False,
                                                 headers={'Content-type': headers.get(content_type,
                                                                                      headers['json'])})
                break
            except Exception:
                log.warning('Request to {} failed. Waiting {} seconds before retrying\n'.format(
                    self.device.name, retry_wait), exc_info=True)
                time.sleep(retry_wait)
        else:
            raise RequestException('Sending "{furl}" to "{d}" has failed '
                                   'after {tries} tries'.format(furl=full_url,
                                                                d=self.device.name,
                                                                tries=retries))

        try:
            # response might not pe in JSON format
            output = response.json()
            log.info("Output received:\n{output}".format(output=output))
        except Exception:
            output = response.content if content_type == 'xml' else response.text
            log.info(f"'Post' operation did not return a json response: {output}")

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}', got:\n {msg}" \
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code,
                                           msg=response.text))
        return output

    @BaseConnection.locked
    @isconnected
    def put(self, api_url, payload=None, expected_status_code=requests.codes.ok,
            content_type='json', timeout=30, retries=3, retry_wait=10):
        """PUT REST Command to configure information from the device
        Arguments
        ---------
            api_url (string): subdirectory part of the API URL
            payload (dict|string): Information to send via the put action
            expected_status_code (int): Expected result
            content_type(string): json / xml / form
            timeout (int): Maximum time
        """

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        # Eliminate the starting "/" if present, as it may cause problems
        api_url = api_url.lstrip('/')
        # Deal with the url
        full_url = '{f}{api_url}'.format(f=self.url, api_url=api_url)

        log.info("Sending PUT command to '{d}':" \
                 "\nURL: {furl}\nPayload:{p}".format(d=self.device.name,
                                                     furl=full_url,
                                                     p=payload))
        headers = {'form': 'application/x-www-form-urlencoded',
                   'json': 'application/json',
                   'xml': 'application/xml'}

        if content_type == 'form' and isinstance(payload, dict):
            payload = urllib.parse.urlencode(payload, safe=':!')
        elif content_type == 'xml':
            if isinstance(payload, dict):
                raise ValueError("Error on {d} during PUT command: "
                                 "Payload must to be string in xml format if used "
                                 "in conjunction with content_type='xml' argument"
                                 .format(d=self.device.name))

        for _ in range(retries):
            try:
                # Send to the device
                if isinstance(payload, dict):
                    response = self.session.put(full_url, json=payload, timeout=timeout,
                                                verify=False)
                else:
                    response = self.session.put(full_url, data=payload, timeout=timeout,
                                                verify=False,
                                                headers={'Content-type': headers.get(content_type,
                                                                                     headers['json'])})
                break
            except Exception:
                log.warning('Request to {} failed. Waiting {} seconds before retrying\n'.format(
                    self.device.name, retry_wait), exc_info=True)
                time.sleep(retry_wait)
        else:
            raise RequestException('Sending "{furl}" to "{d}" has failed '
                                   'after {tries} tries'.format(furl=full_url,
                                                                d=self.device.name,
                                                                tries=retries))

        try:
            # response might not pe in JSON format
            output = response.json()
            log.info("Output received:\n{output}".format(output=output))
        except Exception:
            output = response.content if content_type == 'xml' else response.text
            log.info(f"'Put' operation did not return a json response: {output}")

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}', got:\n {msg}" \
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code,
                                           msg=response.text))
        return output

    @BaseConnection.locked
    @isconnected
    def delete(self, api_url, expected_status_code=requests.codes.ok,
               timeout=30, retries=3, retry_wait=10):
        """DELETE REST Command to delete information from the device
        Arguments
        ---------

            api_url (string): subdirectory part of the API URL
            expected_status_code (int): Expected result
            timeout (int): Maximum time
        """
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        # Eliminate the starting "/" if present, as it may cause problems
        api_url = api_url.lstrip('/')
        # Deal with the url
        full_url = '{f}{api_url}'.format(f=self.url, api_url=api_url)

        log.info("Sending DELETE command to '{d}':" \
                 "\nURL: {furl}".format(d=self.device.name, furl=full_url))

        for i in range(retries):
            try:
                # Send to the device
                response = self.session.delete(full_url, timeout=timeout, verify=False)
                break
            except Exception as e:
                log.warning('Request to {} failed. Waiting {} seconds before retrying\n'.format(
                    self.device.name, retry_wait), exc_info=True)
                time.sleep(retry_wait)
        else:
            raise RequestException('Sending "{furl}" to "{d}" has failed '
                                   'after {tries} tries'.format(furl=full_url,
                                                                d=self.device.name,
                                                                tries=retries))

        try:
            # response might not pe in JSON format
            output = response.json()
            log.info("Output received:\n{output}".format(output=output))
        except ValueError:
            output = response.text
            log.info(f"'Delete' operation did not return a json response: {output}")

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}'" \
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code))

        return output



#!/bin/env python
""" Unit tests for APIC rest.connector """

import os
import unittest
from importlib import import_module

import requests
from requests.models import Response
from unittest.mock import patch, MagicMock
from requests.exceptions import HTTPError

from pyats.topology import loader

from rest.connector import Acisdk
from rest.connector.utils import get_installed_lib_versions
HERE = os.path.dirname(__file__)


class test_rest_connector(unittest.TestCase):

    def setUp(self):
        self.testbed = loader.load(os.path.join(HERE, 'testbed.yaml'))
        self.device = self.testbed.devices['apic']

        libs = get_installed_lib_versions()
        if not all(libs.values()) or len(set(libs.values())) != 1:
            # both libraries should be already installed and have the same version
            self.libs_present = False
        else:
            self.sdk_version = list(libs.values())[0]
            self.libs_present = True

    def test_init(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req().return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
            self.assertEqual(connection.device, self.device)

        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.execute())
        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.configure())

    def test_connection(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
            self.assertEqual(connection.connected, False)

        with patch('requests.post') as req:
            resp = Response()
            resp.status_code = 200
            resp.json = MagicMock(return_value={'imdata': [{'aaaLogin': {
                'attributes': {'token': 'test',
                               'version': 'test',
                               'refreshTimeoutSeconds': 600
                               }
            }}]})
            req.return_value = resp
            connection.connect()
            self.assertEqual(connection.connected, True)
            connection.connect()
            self.assertEqual(connection.connected, True)

        # Now disconnect
        with patch('requests.post') as req:
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_connection_wrong_code(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        self.assertEqual(connection.connected, False)

        with patch('requests.post') as req:
            resp = Response()
            resp.status_code = 404
            req.return_value = resp
            with self.assertRaises(HTTPError):
                connection.connect()

        self.assertEqual(connection.connected, False)

        # Now disconnect
        with patch('requests.post') as req:
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_attribute_call_not_connected(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        with self.assertRaises(Exception):
            connection.lookupByDn('uni')

    def test_attribute_call_connected(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
            self.assertEqual(connection.connected, False)

            with patch('requests.post') as req, patch('requests.Session') as session:
                resp = Response()
                resp.status_code = 200
                resp.json = MagicMock(return_value={'imdata': [{'aaaLogin': {
                    'attributes': {'token': 'test',
                                   'version': 'test',
                                   'refreshTimeoutSeconds': 600
                                   }
                }}]})
                req.return_value = resp
                connection.connect()

                resp._content = str.encode('<?xml version="1.0" ?><imdata totalCount="0"></imdata>')
                session().get.return_value = resp
                mo = connection.create(model='fv.Tenant',
                                       parent_mo_or_dn='uni',
                                       name='test')
                connection.lookupByDn('uni')
                connection.disconnect()
            self.assertEqual(connection.connected, False)

    def test_get_model_wrong_model_format(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')

        with self.assertRaises(NameError):
            connection.get_model(model='wrong_model')

    def test_get_model_wrong_model(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')

        with self.assertRaises(AttributeError):
            connection.get_model(model='fv.Wrong')

    def test_get_model(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        '''
        get_model is not dependant on connection,
        as it leverages the locally installed libraries
        '''
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        self.assertEqual(connection.connected, False)

        model = connection.get_model(model='fv.Tenant')
        tenant_model = getattr(import_module(f'cobra.model.fv'), 'Tenant')
        self.assertEqual(model, tenant_model)

    def test_config_and_commit_not_connected(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        self.assertEqual(connection.connected, False)
        with self.assertRaises(Exception):
            mo = connection.create(model='fv.Tenant',
                                   parent_mo_or_dn='uni',
                                   name='test')

            connection.config_and_commit(mo=mo)

    def test_config_and_commit_connected(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        self.assertEqual(connection.connected, False)

        with patch('requests.post') as req, patch('requests.Session') as session:
            resp = Response()
            resp.status_code = 200
            resp.json = MagicMock(return_value={'imdata': [{'aaaLogin': {
                'attributes': {'token': 'test',
                               'version': 'test',
                               'refreshTimeoutSeconds': 600
                               }
            }}]})
            req.return_value = resp
            connection.connect()

            resp.json = MagicMock(return_value={'imdata': []})
            session().post.return_value = resp
            mo = connection.create(model='fv.Tenant',
                                   parent_mo_or_dn='uni',
                                   name='test')
            connection.config_and_commit(mo=mo)
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_config_and_commit_connected_wrong_status(self):
        if not self.libs_present:
            self.skipTest('Test skipped due to missing libraries')
        with patch('requests.get') as req:
            resp = Response()
            resp.status_code = 200
            resp._content = str.encode('<title>Cisco APIC Python SDK Documentation &#8212; Cisco APIC '
                                       f'Python API {self.sdk_version} documentation</title>')
            req.return_value = resp
            connection = Acisdk(device=self.device, alias='cobra', via='cobra')
        self.assertEqual(connection.connected, False)

        with patch('requests.post') as req, patch('requests.Session') as session:
            resp = Response()
            resp.status_code = 200
            resp.json = MagicMock(return_value={'imdata': [{'aaaLogin': {
                'attributes': {'token': 'test',
                               'version': 'test',
                               'refreshTimeoutSeconds': 600
                               }
            }}]})
            req.return_value = resp
            connection.connect()

            resp2 = Response()
            resp2.status_code = 300
            session().post.return_value = resp2

            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            resp2.json = MagicMock(return_value={'imdata': []})

            with self.assertRaises(Exception):
                mo = connection.create(model='fv.Tenant',
                                       parent_mo_or_dn='uni',
                                       name='test')
                connection.config_and_commit(mo=mo)
            self.assertEqual(connection.connected, True)
            connection.disconnect()
        self.assertEqual(connection.connected, False)


if __name__ == '__main__':
    unittest.main()



# Global imports
import logging
import time

# Genie, pyATS, ROBOT imports
# from pyats.connections import BaseConnection
from rest.connector.utils import get_username_password
from rest.connector.implementation import Implementation
from pyats.connections import BaseConnection

# F5 imports
from icontrol.session import iControlRESTSession
from icontrol.exceptions import iControlUnexpectedHTTPError

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Implementation):

    """Rest BaseClass



    Baseclass for Rest connection implementation



    YAML Example

    ------------



        devices:

            bigip1:
                type: 'generic'
                os: 'bigip'
                custom:
                    abstraction:
                        order: [os]
                connections:
                    # Console
                    a:
                        ip: 1.2.3.4
                        port: 22
                        protocol: ssh
                    rest:
                        # specify the rest connector class
                        class: rest.connector.Rest
                        ip: 1.2.3.4
                        port: "443"
                        protocol: https
                        credentials:
                            rest:
                                username: user
                                password: password



    Example

    -------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/path/to/testbed.yaml')
        >>> device = testbed.devices['bigip1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected

        True

    """

    @property
    def connected(self):

        """Is a device connected"""

        return self._is_connected

    def disconnect(self):
        """disconnect the device for this particular alias"""
        if self.token:
            try:
                log.info("Deleting token: '%s'", self.token)
                delete_url = (
                    f"{self.base_url}/mgmt/shared/authz/tokens/{self.token}"
                )

                delete_icr_session = iControlRESTSession(
                    self.username,
                    self.password,
                    verify=self.verify,
                    token_to_use=self.token
                )
                # Revoking the token received
                response = delete_icr_session.delete(delete_url)

                if not response.ok or response.status_code != 200:
                    log.error(
                        "Failed to delete session token for device %s",
                        self.device.name
                    )
                else:
                    log.info(
                        "Session token for device %s successfully deleted",
                        self.device.name
                    )
            finally:
                self.token = None
                self._is_connected = False

    def isconnected(func):
        '''Decorator to make sure the session to device is active.
        If the token experied, it will attempt to reconnect
        '''

        def decorated(self, *args, **kwargs):
            try:
                result = func(self, *args, **kwargs)
            except iControlUnexpectedHTTPError as ex:
                # Auth failure - probably token expired
                if ex.response.status_code == 401:
                    log.info("Session with device %s expired", self.device.name)
                    log.info("Reconnecting to device %s", self.device.name)
                    self._is_connected = False
                    timeout = kwargs['timeout'] if 'timeout' in kwargs else 30
                    self._connect(timeout, retries=0, retry_wait=0)
                    result = func(self, *args, **kwargs)
                else:
                    raise
            return result
        return decorated

    def connect(
        self,
        auth_provider='tmos',
        verify=False,
        port='443',
        protocol='https',
        retries=3,
        retry_wait=10,
        timeout=30,
        ttl=3600,
        *args,
        **kwargs
    ):
        if self.connected:
            return

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    "`unicon` is not installed for `sshtunnel`. "
                    "Please install by `pip install unicon`."
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    self.ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    self.port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            self.ip = self.connection_info['ip'].exploded
            self.port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.base_url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=self.ip,
                                                          port=self.port)

        self.username, self.password = get_username_password(self)

        self.header = "Content-Type: application/json"
        self.verify = verify
        self._auth_provider = auth_provider
        self._ttl = ttl

        self._connect(timeout, retries, retry_wait)

        return self._is_connected, self.icr_session

    def _connect(self, timeout: int, retries: int, retry_wait: int):
        """ Authenticate and initiate a session with the device

        Args:
            timeout: The timeout to use when establishing the connection
            retries: How many times to retry to connect to the device if it fails
            retry_wait: Time in seconds to wait between retries
        """
        self._authenticate(timeout, retries, retry_wait)

        self._extend_session_ttl(self._ttl)

        params = dict(
            username=self.username,
            password=self.password,
            verify=self.verify,
            token_to_use=self.token
        )

        # creating an object to be used all new requests
        self.icr_session = iControlRESTSession(**params)

        self._is_connected = True

        log.info(
            "Connected successfully to '%s'", self.device.name
        )

    def _authenticate(self, timeout: int, retries: int, retry_wait: int):
        """ Authenticates with the device and retrieves a session token to be
            used in actual requests

        Args:
            timeout: The timeout to use when establishing the connection
            retries: How many times to retry to connect to the device if it fails
            retry_wait: Time in seconds to wait between retries
        """
        # URL to authenticate and receive the token
        url = f"{self.base_url}/mgmt/shared/authn/login"

        payload = {
            'username': self.username,
            'password': self.password,
            'loginProviderName': self._auth_provider
        }

        iCRS = iControlRESTSession(
            self.username,
            self.password,
            timeout=timeout,
            verify=self.verify
        )

        log.info(
            "Connecting to '%s'", self.device.name
        )

        response = iCRS.post(
            url,
            json=payload,
        )

        log.debug(response.json())

        if response.status_code != 200:
            if b'Configuration Utility restarting...' in response.content:
                if retries > 0:
                    time.sleep(retry_wait)
                    return self._authenticate(timeout, retries - 1, retry_wait)
                else:
                    raise iControlUnexpectedHTTPError(
                        f"Failed to connect to {self.device.name}: "
                        f"{response.content}"
                    )
            else:
                raise iControlUnexpectedHTTPError(
                    f"Failed to authenticate with {self.device.name}"
                )

        self.token = response.json()['token']['token']

        log.debug(
            "The following token is used to connect: '%s'", self.token
        )

    def _extend_session_ttl(self, ttl: int) -> None:
        """ Sets the TTL for the active session

        Args:
            ttl: The TTL to be set for the session
        """
        # Self-link of the token
        timeout_url = f"{self.base_url}/mgmt/shared/authz/tokens/{self.token}"
        timeout_payload = {"timeout": ttl}
        token_icr_session = iControlRESTSession(
            self.username,
            self.password,
            verify=self.verify,
            token_to_use=self.token
        )
        # Extending the timeout for the token received
        response = token_icr_session.patch(timeout_url, json=timeout_payload)
        if response.status_code != 200 or not response.ok:
            raise iControlUnexpectedHTTPError(
                "Failed to refresh session: "
                f"{response.reason} ({response.status_code})"
            )
        log.debug("Token TTL extended to '%d' seconds", ttl)

    @isconnected
    def get(self, api_url, timeout=30, verbose=False):

        """GET REST Command to retrieve information from the device"""

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending GET to '{d}': "
            "{u}".format(d=self.device.name, u=full_url)
        )

        response = self.icr_session.get(full_url, timeout=timeout)

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=response))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        log.debug(
            "Successfully fetched data using token: '{t}'".format(t=self.token)
        )

        return response

    @BaseConnection.locked
    @isconnected
    def post(self, api_url, payload, timeout=30, verbose=False):
        """POST REST Command to configure information from the device"""

        if not self.connected:
            raise Exception(
                f"No active connection for '{self.device.name}'"
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.post(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def put(self, api_url, payload, timeout=30, verbose=False):

        """PUT REST Command to update information on the device"""

        if not self.connected:
            raise Exception(
                f"No active connection for '{self.device.name}'"
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.put(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def patch(self, api_url, payload, timeout=30, verbose=False):
        """PATCH REST Command to update information on the device"""

        if not self.connected:
            raise Exception(
                f"No active connection for '{self.device.name}'"
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.patch(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def delete(self, api_url, timeout=30, verbose=False):

        """DELETE REST Command to delete information from the device"""

        if not self.connected:
            raise Exception(
                f"No active connection for '{self.device.name}'"
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'".format(
                d=self.device.name, u=full_url, h=self.header
            )
        )

        response = self.icr_session.delete(full_url, timeout=timeout)

        output = response.text

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    def configure(self, *args, **kwargs):

        """configure - Not implemented for REST"""

        raise NotImplementedError(
            "configure is not a supported method for REST. "
            "post is probably what you are looking for"
        )

    def execute(self, *args, **kwargs):

        """execute - Not implemented for REST"""

        raise NotImplementedError(
            "execute is not a supported method for REST. "
            "get is probably what you are looking for."
        )



#! /usr/bin/env python

"""Setup file for rest package

See:
    https://packaging.python.org/en/latest/distributing.html
"""

import os
import re
import sys
import shlex
import unittest
import subprocess
from setuptools import setup, find_packages, Command
from setuptools.command.test import test

pkg_name = 'rest.connector'
pkg_path = '/'.join(pkg_name.split('.'))


class CleanCommand(Command):
    '''Custom clean command

    cleanup current directory:
        - removes build/
        - removes src/*.egg-info
        - removes *.pyc and __pycache__ recursively

    Example
    -------
        python setup.py clean

    '''

    user_options = []
    description = 'CISCO SHARED : Clean all build artifacts'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./src/*.egg-info')
        os.system('find . -type f -name "*.pyc" | xargs rm -vrf')
        os.system('find . -type d -name "__pycache__" | xargs rm -vrf')


class TestCommand(Command):
    user_options = []
    description = 'CISCO SHARED : Run unit tests against this package'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # where the tests are (relative to here)
        tests = os.path.join('src', pkg_path, 'tests')

        # call unittests
        sys.exit(unittest.main(
            module=None,
            argv=['python -m unittest', 'discover', tests],
            failfast=True))


def read(*paths):
    '''read and return txt content of file'''
    with open(os.path.join(os.path.dirname(__file__), *paths)) as fp:
        return fp.read()


def find_version(*paths):
    '''reads a file and returns the defined __version__ value'''
    version_match = re.search(r"^__version__ ?= ?['\"]([^'\"]*)['\"]",
                              read(*paths), re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# launch setup
setup(
    name=pkg_name,
    version=find_version('src', pkg_path, '__init__.py'),

    # descriptions
    description='pyATS REST connection package',
    long_description=read('DESCRIPTION.rst'),

    # the package's documentation page.
    url='https://developer.cisco.com/docs/rest-connector/',

    # author details
    author='Cisco Systems Inc.',
    author_email='jeaubin@cisco.com',
    maintainer_email='pyats-support-ext@cisco.com',

    # project licensing
    license='Apache 2.0',

    platforms=['Linux', 'macOS'],

    # see https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 6 - Mature',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    # uses namespace package
    namespace_packages=['rest'],

    # project keywords
    keywords='pyats cisco-shared',

    # project packages
    packages=find_packages(where='src'),

    # project directory
    package_dir={
        '': 'src',
    },

    # additional package data files that goes into the package itself
    package_data={'': ['README.md',
                       'tests/testbed.yaml',
                       ]
                  },

    # Standalone scripts
    scripts=[
    ],

    # console entry point
    entry_points={
    },

    # package dependencies
    install_requires=[
        'requests >= 1.15.1',
        'dict2xml',
        'f5-icontrol-rest',
    ],

    # any additional groups of dependencies.
    # install using: $ pip install -e .[dev]
    extras_require={
        'dev': ['coverage',
                'restview',
                'Sphinx',
                'sphinx-rtd-theme',
                'requests-mock'],
    },

    # any data files placed outside this package.
    # See: http://docs.python.org/3.4/distutils/setupscript.html
    # format:
    #   [('target', ['list', 'of', 'files'])]
    # where target is sys.prefix/<target>
    data_files=[],

    # custom commands for setup.py
    cmdclass={
        'clean': CleanCommand,
        'test': TestCommand,
    },

    # non zip-safe (never tested it)
    zip_safe=False,
)
